/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * The contents of this file are subject to the Mozilla Public 
 * License version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or 
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Initial Developer of the Original Code is R.V.Guha. Portions
 * created by R.V.Guha are copyright (c) 2000 R.V.Guha. All Rights
 * Reserved.
 *
 * Contributors : Eric Van Der Vlist, Ivan Kurmanov, Jasper van de Gronde (win32 port)
 */


#include "../utils/utils.h"
#include "queryParser.h"
#include "selectParser.h"
#include "../regexp/regexp.h"
#include <stdio.h>
#ifdef WIN32
#include <time.h>
#endif

extern int       DB_Flush (const char* db) ;
extern int gVerbose;
extern int gAssertionCount;
void addNamespace (ClientConnection cl, char *prefix, char* url);
void removeNamespace (ClientConnection cl, char* prefix);

#define VALUE_BUF 4048

#define DEBUG_PROCESS_QUERY 0

#if DEBUG_PROCESS_QUERY
#define DEBUG if(1)
#else
#define DEBUG if(0)
#endif

/* Jasper van de Gronde
 * Holds the global namespace mappings.
 */
NamespaceStruct* gNamespaceList = 0;

/* Jasper van de Gronde
 * This initializes anything processQuery might need.
 * At the moment it just initializes the global namespace list.
 */
void
InitializeProcessQuery() {
  const char* const namespaces = getenv("RDFDB_NAMESPACES");
  if ( namespaces ) {
    NamespaceStruct* curnamespace = gNamespaceList;
    const char* curprefixbegin = 0;
    const char* curprefixend = 0;
    const char* curnsbegin = 0;
    const char* curpos;
    for(curpos = namespaces; *curpos; ++curpos) {
      if ( !curprefixbegin ) {
        curprefixbegin = curpos;
      }
      else if ( !curprefixend ) {
        if ( *curpos == '<' ) {
          curprefixend = curpos;
          curnsbegin = curpos+1;
        }
      }
      else {
        if ( *curpos == '>' ) {
          if ( !curnamespace ) {
            gNamespaceList = (NamespaceStruct*)getMem(sizeof(NamespaceStruct));
            curnamespace = gNamespaceList;
          }
          else {
            curnamespace->next = (NamespaceStruct*)getMem(sizeof(NamespaceStruct));
            curnamespace = curnamespace->next;
          }
          if ( !curnamespace ) break;
          curnamespace->prefix = strduprange(curprefixbegin, curprefixend);
          curnamespace->url = strduprange(curnsbegin, curpos);

          curprefixbegin = 0;
          curprefixend = 0;
          curnsbegin = 0;
        }
      }
    }
  }
}

/* Jasper van de Gronde
 * This deinitializes anything InitializeProcessQuery might have initialized.
 */
void
DeinitializeProcessQuery() {
  NamespaceStruct* next;
  while(gNamespaceList) {
    if ( gNamespaceList->prefix ) freeMem(gNamespaceList->prefix);
    if ( gNamespaceList->url ) freeMem(gNamespaceList->url);
    next = gNamespaceList->next;
    freeMem((char*)gNamespaceList);
    gNamespaceList = next;
  }
}

/* Jasper van de Gronde
 * This searches for the variable named var.
 * If the variable isn't found it returns 0.
 * NOTE: This isn't used at the moment.
 */
Variable
searchVariable(const Query q, const char* const var) {
  Variable v = q->variables;
  while (v) {
    if (streq(v->name, var)) {
      break;
    }
    v = v->next;
  }
  return v;
}


int 
varp (Query q, const char* var) {
  if (var[0] != '?') {
    return 0;
  } else {
    Variable v = q->variables;
    while (v) {
      if (streq(v->name, var)) {
        return v->value == 0;
      }
      v = v->next;
    }
  }
  return 1;
}

const char*
vval (Query q, const char* var) {
  if (var[0] != '?') {
    return var;
  } else {    
    Variable v = q->variables;
    while (v) {
      if (streq(v->name, var)) {
        return v->value;
      }
      v = v->next;
    }
    return var;
  }
}
  
RDF_ValueType
vtype (Query q, const char* var) {
  if (var[0] != '?') {
    if (var[0] == '\'') {
      return RDF_STRING_TYPE;
    } else return RDF_RESOURCE_TYPE;
  } else {    
    Variable v = q->variables;
    while (v) {
      if (streq(v->name, var)) {
        return v->type;
      }
      v = v->next;
    }
    return 0;
  }
}

    
void
setVariableValue  (Query q, const char* var, void* value, RDF_ValueType type) {
  Variable v = q->variables;
  while (v) {
    if (streq(v->name, var)) {
      v->value = value;
      v->type  = type;
      break;
    }
    v = v->next;
  }
}


/* Jasper van de Gronde
 * This checks wether the var-value-type combination isn't in conflict with the constraint list.
 */
BOOL
checkConstraintList(Query q, const char* var, void* value, RDF_ValueType type) {
  struct constraint_list_t* constraint;

  // this is done because the old code left the '?' in front of the variable name,
  // while the new code doesn't (so ?x in the old code is just x in the new code)
  if ( var && *var ) var += 1;

  constraint = q->constraint_list;
  while(constraint) {
    if ( streq(var, constraint->expr.var) ) {
      if ( constraint->expr.var_or_literal.type != type ) return FALSE;
      switch(constraint->expr.operation) {
      case STRCOMP_LIKE:
        if ( !regexec(constraint->expr.rx,(char*)value) ) return FALSE;
        break;
      case STRCOMP_NE:
        if ( streq((char*)value,constraint->expr.var_or_literal.text) ) return FALSE;
        break;
      case STRCOMP_EQ:
        if ( !streq((char*)value,constraint->expr.var_or_literal.text) ) return FALSE;
        break;
      case STRCOMP_SOMETHING: // still have to find out what this does ('~')
        // TODO: Implement this;
        break;

      case NUMCOMP_GT: // '>'
        if ( (int)value <= constraint->expr.var_or_literal.num ) return FALSE;
        break;
      case NUMCOMP_LT: // '<'
        if ( (int)value >= constraint->expr.var_or_literal.num ) return FALSE;
        break;
      case NUMCOMP_EQ: // both '==' and '='
        if ( (int)value != constraint->expr.var_or_literal.num ) return FALSE;
        break;
      case NUMCOMP_NE: // both '!=' and '<>'
        if ( (int)value == constraint->expr.var_or_literal.num ) return FALSE;
        break;
      case NUMCOMP_LTEQ: // '<='
        if ( (int)value > constraint->expr.var_or_literal.num ) return FALSE;
        break;
      case NUMCOMP_GTEQ: // '>='
        if ( (int)value < constraint->expr.var_or_literal.num ) return FALSE;
        break;

      default:
        // Shouldn't occur as this should have triggered a syntax error.
        return FALSE;
      }
    }
    constraint = constraint->next;
  }
  return TRUE;
}


void
outputEscapedString (char* str, char* line) {
  // str - source, line - destination buffer pointer

  char* vs ;
  //  *line++ = '\'';    // IKu
  for (vs = strchr(str, '\''); vs != 0; vs = strchr(str, '\'')) {
    memcpy(line, str, vs-str);
    line += vs-str;
    *line++ = '\\';
    *line++ = '\'';
    str = vs+1;
  }
  memcpy(line, str, strlen(str));
  line += strlen(str);
  //  *line++ = '\''; // IKu fixed, was: ... = '\\'; 
  //  *line++ = '\t'; // IKu
  *line = 0;
}

int
addAnswer (Query q) {
  Variable v = q->variables;
//  int l;

  if (q->outputType == TAB_LIMITED_OUTPUT && !q->queryHeaderPrinted) {
    Variable v1 = q->variables;
    q->queryHeaderPrinted = 1;
    while (v1) {
      if (v1->outputVariablep) { // used to be if (v->outputVariablep) {
        writen(q->cl->socket, v1->name, strlen(v1->name));
        writen(q->cl->socket, "\t", 1);
      }
      v1 = v1->next;
    }
    if (q->cl->socket)  writen(q->cl->socket, NETWORK_NEWLINE, NETWORK_NEWLINE_LEN);
  }

  while (v) {
    if (v->value && v->outputVariablep) {
      char ansLine[VALUE_BUF];

      if( strlen( (char*)v->value ) > VALUE_BUF-30 ) { 
        /*** this is an exception and is a good reason to complain and die ***/
      }

      if (v->type == RDF_RESOURCE_TYPE) {
        if (q->outputType == TAB_LIMITED_OUTPUT) {
          snprintf(ansLine, VALUE_BUF, "<%s>\t", (char*)v->value);
        } else {
          snprintf(ansLine, VALUE_BUF, "%s = <%s>\t", v->name, (char*)v->value);
        }
      } else if (v->type == RDF_STRING_TYPE) {

        char* destination = ansLine;

        /**  value start  **/ 
        if (q->outputType == TAB_LIMITED_OUTPUT) {
          (*destination) = '\'';
          destination++;
        } else { 
          snprintf( destination, VALUE_BUF, "%s = '", v->name );
          destination += strlen( destination );
        }

        /**  here goes the value itself  **/
        if ( (strchr((char*)v->value, '\''))
   //             || (strchr((char*)v->value, '\t')) 
   //             || (strchr((char*)v->value, NEWLINE_CHAR)) 
             ) {
          outputEscapedString((char*)v->value, destination);
        } else {
          strcpy( destination, (char*)v->value );
        }
        destination += strlen( destination );

        (*destination++) = '\'';
        (*destination++) = '\t';
        (*destination++) = 0;
        
      } else if (v->type == RDF_INT_TYPE) {
        if (q->outputType == TAB_LIMITED_OUTPUT) 
          snprintf(ansLine, VALUE_BUF, "%i\t", (int)v->value);
        else 
          snprintf(ansLine, VALUE_BUF, "%s = %i\t", v->name, (int)v->value);
      }
      if (q->cl->socket) {
        writen(q->cl->socket, ansLine, strlen(ansLine));
      } else return 0;
      if (gVerbose > 0) {
        printf("%s", ansLine);
      }
    }
    v = v->next;
  }
  if (gVerbose > 0) {
    printf("\n");
  }
  if (q->cl->socket)  writen(q->cl->socket, NETWORK_NEWLINE, NETWORK_NEWLINE_LEN);
  return 1;
}


/**  
     recursive Select function:
       int SelectNextInt 
     parameters: 
       Query q, 
       triple Literal nl 
     returns:
       0 upon normal finish;
       !0 unon error (error code);
     
**/
int
SelectNextInt (Query q, Literal nl) {
  int err;
  if (q->abort) return 0;

  if (!nl) {

    if( q->outputType > 0 ) {
      /* we must print out what we found */ 
      if (!addAnswer(q)) {
        q->abort = 1;
        return 0;
      }
    }

    //    DEBUG printf ( "-query outputType: %i\n", q->outputType );
    //    DEBUG printf ( "-after procesing a valid assertion\n" );

  } else if (!varp(q, nl->arc) && !varp(q, nl->source) && !varp(q, nl->target)) {
    /* 
       that means: if all three elements of the triple are already exact values, 
       either specified by the user in the original query, 
       or found as satisfying to the original query condition 
    */
    
    /* 
       just checking ... and recursing to check that if would satisfy other 
       conditions as well 
    */
    if (0 != (err = RDF_HasAssertion(q->database, (const RDF_Resource) vval(q, nl->source), 
                                      (const RDF_Resource) vval(q, nl->arc), 
                                     vval(q, nl->target), vtype(q, nl->target)))) {

      if (err < 0) return err;

      SelectNextInt(q, nl->nxt);
    }

  } else {  /*  at least one of the triple elements is a variable  */ 

    /***  
          therefore we need to search for data triples, that satisfy
          the condition requirement.  We do this by calling
          RDF_Get... (..., &c ) function, which would get us the data
          in the RDF_Cursor c;
          
          Also, if in the current triple there are two parameter variables, then we will 
          have to cycle through all the first variable's values' first, looking for possible
          second variable values at the same time....
    ***/

    RDF_Cursor c = 0;
    Literal next;
    int len;
    const char* var;

    if ((!varp(q, nl->arc)) && varp(q, nl->source) && (!varp(q, nl->target))) {

      /*** source (resource) is a variable ***/
      var = nl->source;
      next = nl->nxt;
      err = RDF_GetSources(q->database, 
                           (const RDF_Resource) vval(q, nl->target), 
                           (const RDF_Resource) vval(q, nl->arc), &c);

    } else if ((!varp(q, nl->arc)) && !varp(q, nl->source) && (varp(q, nl->target))) {

      /*** target (object) is a variable ***/
      var = nl->target;
      next = nl->nxt;
      err = RDF_GetTargets(q->database, 
                           (const RDF_Resource) vval(q, nl->source), 
                           (const RDF_Resource) vval(q, nl->arc),  &c);

    } else if ((!varp(q, nl->arc)) && varp(q, nl->source) && varp(q, nl->target)) {
      /*** source (resource) and target are variables ***/
      var = nl->source;
      next = nl;
      err = RDF_GetArcSources(q->database, (const RDF_Resource) vval(q, nl->arc),  &c);

    } else if (varp(q, nl->arc) && (!varp(q, nl->source))) {
      /***  arc (property) is a variable, but the source (subject) is not  ***/
      var = nl->arc;
      next = (varp(q, nl->target) ? nl : nl->nxt);
      err = RDF_GetArcsOut(q->database, (const RDF_Resource) vval(q, nl->source),  &c);

    } else if (varp(q, nl->arc) && (!varp(q, nl->target))) {
      /***  arc (property) is a variable, but the target (object) is not  ***/
      var = nl->arc;
      next = (varp(q, nl->source) ? nl : nl->nxt);
      err = RDF_GetArcsIn(q->database, (const RDF_Resource) vval(q, nl->target),  &c);

    } else {
      /*  all three triple elements are unconstrained variables   */     
      return RDF_ERROR;
    }

    DEBUG 
      printf ( "\n-starting cursor: %x\n", c );
          
    if (err) return err;
    if (!c) return RDF_ERROR;
    
    while (0 == RDF_Advance(c)) {      
      void *value;
      RDF_ValueType type;

      DEBUG printf ( "-(another) item found in cursor %x\n", c );

      RDF_CursorCurrentValue(c, &value);
      RDF_CursorCurrentValueSize(c, &len);
      RDF_CursorCurrentValueType(c, &type);
      if (type == RDF_RESOURCE_TYPE || type == RDF_STRING_TYPE) ((char*)value)[len] = 0; 

      DEBUG printf ( "-variable %s's value is '%s'\n", var, value );

      /* Jasper van de Gronde
       * Added the checkConstraint bit
       */
      if ( checkConstraintList(q, var, value, type) ) {
        setVariableValue(q, var, value, type);

        /**** delete if necessary ****/
        if( (q->type == DELETEWHERE) && (!next) ) {
          char * src, *arc, *target;
          int type;
          int err;
        
          DEBUG printf ( "-deleting an assertion\n" );
        
          src    = (char*) vval( q, nl->source );
          arc    = (char*) vval( q, nl->arc    );
          target = (char*) vval( q, nl->target );
          type   = (int)   vtype(q, nl->target );
        
          DEBUG printf ( "'%s' = %s => '%s' (type: %i)\n", 
                         src, arc, target, type
                         ) ;
        
          err = RDF_Unassert( q->database, 
                              (const RDF_Resource) src, 
                              (const RDF_Resource) arc, 
                              target, type ) ;
        
          //        if (err < 0) return err;
        }

        err = SelectNextInt(q, next);

        if (err) {
          RDF_DisposeCursor(c);
          return err;
        }
      }

    }
    DEBUG printf ( "-no more items in cursor\n" );

    setVariableValue(q, var, 0, 0);
    RDF_DisposeCursor(c);  
  }

  DEBUG printf ( "-quitting SelectNextInt normally\n" );

  return 0;
}


char*
appendNamespace (const char* str, Namespace ns) {
 char* st =  strchr(str, ':');
 if (!st) {
   if (strlen(ns->prefix) > 0) {
     return strdup(str);
   } else {
     /* Jasper van de Gronde
      * Fixed the memory allocation (it used to allocate one byte too little).
      */
     char* nn =  getMem(strlen(ns->url) + 1 + strlen(str) + 1);
     memcpy(nn, ns->url, strlen(ns->url));
     strcat(nn, "#");
     strcat(nn, str);
     return nn;
   }
 } else {
   /* Jasper van de Gronde
    * Fixed the memory allocation (it used to allocate one byte too little).
    */
   char* nn =  getMem(strlen(ns->url) + 1 + strlen(st+1) + 1);
   memcpy(nn, ns->url, strlen(ns->url));
   strcat(nn, "#");
   strcat(nn, st+1);
   return nn;
 }
}

int
namespaceMatch(const char* uri, Namespace ns) {
  if (strlen(ns->prefix) == 0) {
    return (strchr(uri, ':') == 0);
  } else {
    const char* st = strchr(uri, ':');
    if (!st) return 0;
    if ((size_t)(st-uri) != strlen(ns->prefix)) return 0;
    return (strncmp(uri, ns->prefix, st-uri) == 0);
  }
}

void
namespaceMungeLiteral (Namespace ns, Literal l) {
  if (l->source[0] != '?') {
    Namespace nsx = ns;
    while (nsx) {
      if (namespaceMatch(l->source, nsx)) {
        const char *nn = appendNamespace(l->source, nsx);
        freeMem(l->source);
        l->source = nn;
        break;
      }
      nsx = nsx->next;
    }
  }
  if (l->type == RDF_RESOURCE_TYPE) {
    char* trg = (char*) l->target;
    if (trg[0] != '?') {
      Namespace nsx = ns;
      while (nsx) {
        if (namespaceMatch(trg, nsx)) {
          char *nn = appendNamespace(trg, nsx);
          freeMem(l->target);
          l->target = nn;
          break;
        }
        nsx = nsx->next;
      }
    }        
  }
  {
    char* arc = (char*) l->arc;
    Namespace nsx = ns;
    while (nsx) {
      if (namespaceMatch(arc, nsx)) {
        char *nn = appendNamespace(arc, nsx);
        freeMem(l->arc);
        l->arc = nn;
        break;
      }
      nsx = nsx->next;
    }
  }        
}


/* Jasper van de Gronde
 * Does practically the same as namespaceMungeLiteral, only for a constraint.
 */
void
namespaceMungeConstraint (Namespace ns, struct constraint_list_t* l) {
  if (l->expr.var_or_literal.type == RDF_RESOURCE_TYPE) {
    char* resource = l->expr.var_or_literal.text;
    Namespace nsx = ns;
    while (nsx) {
      if (namespaceMatch(resource, nsx)) {
        char *nn = appendNamespace(resource, nsx);
        freeMem(l->expr.var_or_literal.text);
        l->expr.var_or_literal.text = nn;
        break;
      }
      nsx = nsx->next;
    }
  }
}


Literal
orderLiterals1 (Query q) {
  Literal l, n;
  n = 0;
  for (l = q->literals; l != 0; l = l->next) {
    l->score = 0;
    if (varp(q, l->arc)) l->score++;
    if (varp(q, l->source)) l->score++;
    if (varp(q, l->target)) l->score++;
  }
  for (l = q->literals; l != 0; l = l->next) {
    if (!l->taken) {
      n = l; break;
    }
  }
  for (l = q->literals; l != 0; l = l->next) {
    if (!l->taken && l->score < n->score) 
      n = l;
  }
  return n;
}

/**** 
      reorder Literals function: void orderLiterals() 

      the triples of the query's where statement are reordered by this
      function and some preparational processing is being done at this
      level

      new order of the triples is saved in the literal->nxt, instead
      of literal->next original by the queryParser
*****/

void
orderLiterals (Query q) {
  Literal  next, prev;
  for (next =  orderLiterals1(q); next != 0; next = orderLiterals1(q)) {
    if (!q->lits) {
      q->lits = next;
    } else {
      prev->nxt = next;
    }
    if (varp(q, next->source)) {
      setVariableValue(q, next->source, "temp", RDF_STRING_TYPE);
    }
    if (varp(q, next->target)) {
      setVariableValue(q, next->target, "temp", RDF_STRING_TYPE);
    }
    next->taken = 1;
    prev = next;
  }    
  {
    Variable v ;
    for (v = q->variables; v != 0; v = v->next) {
      v->value = 0;
    }
  }
}


int
processQueryInt (const char* queryString, Query *qpt, ClientConnection cl) {
  QueryParseResult err;
  Query            query;
  Literal l ;
  struct constraint_list_t* c;

  if (gVerbose > 0) {
    time_t tm ;
    tm = time(&tm);
    printf("%sQuery> %s\n", ctime(&tm), queryString);
    fflush(stdout);
  }


  if (0 != (err = parseQuery(queryString, qpt))) {
    return err;
  }


  query = *qpt;

  query->cl = cl;

  /* Jasper van de Gronde
   * Added the namespaceMungeConstraint and gNamespaceList bits.
   */
  if (query->for_list) {
    for (l = query->literals; l != 0; l = l->next) {
      namespaceMungeLiteral(query->for_list, l);
    }
    for (c = query->constraint_list; c != 0; c = c->next) {
      namespaceMungeConstraint(query->for_list, c);
    }
  }
  if (cl->namespaces) {
    for (l = query->literals; l != 0; l = l->next) {
      namespaceMungeLiteral(cl->namespaces, l);
    }
    for (c = query->constraint_list; c != 0; c = c->next) {
      namespaceMungeConstraint(cl->namespaces, c);
    }
  }
  if (gNamespaceList) {
    for (l = query->literals; l != 0; l = l->next) {
      namespaceMungeLiteral(gNamespaceList, l);
    }
    for (c = query->constraint_list; c != 0; c = c->next) {
      namespaceMungeConstraint(gNamespaceList, c);
    }
  }


  if (gVerbose > 2) {
    describeQuery (query) ;
  }

  switch (query->type) {

  case DELETEWHERE :    
  case SELECT :
    {
      orderLiterals(query);
      if (0 != (err = SelectNextInt(query, query->lits))) {
        freeQuery(query);
        query = 0;
        return err;
      }
      break;
    }

  case UPLOAD :
    {
      char* code = 0;
      int start = gAssertionCount;
      err = RDF_LoadURLIntoDB (query->database, query->url, query->fileType, 1, &code);
      DB_Flush (query->database) ;
      printf("%i Assertions Added\n", gAssertionCount-start);
      if (code) {
        writen(cl->socket, code, strlen(code));
        writen(cl->socket, NETWORK_NEWLINE, NETWORK_NEWLINE_LEN);
      }
      break;
    }

  case UNLOAD :
    {
      err = RDF_UnLoadURLFromDB (query->database, query->url);
      DB_Flush (query->database) ;
      break;
    }

  case CREATE :
    {
      RDF_DB db;
      err = RDF_GetDB (query->database, 1, &db);
      break;
    }
    
  case INSERT :
    {
      Literal l = query->literals;
      while (l) {
        err = RDF_Assert(query->database, l->source, l->arc, l->target, l->type);
        l = l->next;
      }
      DB_Flush (query->database) ;
      break;
    }

  case DELETE :    
    {
      Literal l = query->literals;
      while (l) {
        err = RDF_Unassert(query->database, l->source, l->arc, l->target, l->type);
        l = l->next;
      }
      DB_Flush (query->database) ;
      break;
    }

  case ENTER :    
    {
      char* prefix = query->namespacePrefix;
      if (startsWith(prefix, "xmlns:")) {
        addNamespace(cl, &prefix[6], query->namespace);
      } else {
        addNamespace(cl, "", query->namespace);
      }
      break;
    }

  case LEAVE :    
    {
      char* prefix = query->namespacePrefix;
      if (startsWith(prefix, "xmlns:")) {
        removeNamespace(cl, &prefix[6]/*, query->namespace*/);
      } else {
        removeNamespace(cl, ""/*, query->namespace*/);
      }
      break;
    }

  }
  
  /*   write(fd, "%i", q->id); */
  if (err) {
    return err;
  } else {
    return query->id;
  }
}

#define HELP_MSG "Please look at http://web1.guha.com/rdfdb/ for help"

int
ProcessQuery (const char* query, ClientConnection cl) {
#ifdef WIN32
#else
  int fd = cl->socket;
#endif
  Query q = 0;
  int qid;
  char errStr[32];

  if (streq(query, "help")) {
#ifdef WIN32
    writen(cl->socket, HELP_MSG, strlen(HELP_MSG));
#else
    writen(fd, HELP_MSG, strlen(HELP_MSG));
#endif
    return 0;
  }

  qid = processQueryInt(query, &q, cl);

  if (cl->socket) {
    if (qid < 0) {
      snprintf(errStr, 32, "%i </>" NETWORK_NEWLINE, qid);
#ifdef WIN32
    writen(cl->socket, errStr, strlen(errStr));
#else
    writen(fd, errStr, strlen(errStr));
#endif
    } else {
#ifdef WIN32
    writen(cl->socket, "0 </>" NETWORK_NEWLINE, strlen("0 </>" NETWORK_NEWLINE));
#else
    writen(fd, "0 </>" NETWORK_NEWLINE, strlen("0 </>" NETWORK_NEWLINE));
#endif
    }  
  }

  if (q && !qid) freeQuery(q);
  
  return 0;
}
