/* Copyright (c) Jasper van de Gronde, 2004
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above
 *     copyright notice, this list of conditions and the
 *     following disclaimer.
 *  2. Redistributions in binary form must reproduce the
 *     above copyright notice, this list of conditions
 *     and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or
 *     promote products derived from this software without
 *     specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "selectParser.h"
#include "../utils/utils.h"

char* duplicateStr(const char* start, const char* end);

typedef enum {
  ID_NONE=-1,
  ID_SELECT=0,
  ID_FROM,
  ID_WHERE,
  ID_AND,
  ID_USING,
  ID_FOR,

  ID_NUMCOMP_GT, // '>'
  ID_NUMCOMP_LT, // '<'
  ID_NUMCOMP_EQ1, // '=='
  ID_NUMCOMP_EQ2, // '='
  ID_NUMCOMP_NE1, // '!='
  ID_NUMCOMP_NE2, // '<>'
  ID_NUMCOMP_LTEQ, // '<='
  ID_NUMCOMP_GTEQ, // '>='

  ID_STRCOMP_LIKE,
  ID_STRCOMP_NE,
  ID_STRCOMP_EQ,
  ID_STRCOMP_SOMETHING, // still have to find out what this does ('~')
} id_type;
const char* const ids[] = {
  "select",
  "from",
  "where",
  "and",
  "using",
  "for",

  ">",
  "<",
  "==",
  "=",
  "!=",
  "<>",
  "<=",
  ">=",

  "like",
  "ne",
  "eq",
  "~",
};
const size_t id_lengths[] = {
  6, // select
  4, // from
  5, // where
  3, // and
  5, // using
  3, // for

  1, // >
  1, // <
  2, // ==
  1, // =
  2, // !=
  2, // <>
  2, // <=
  2, // >=

  4, // like
  2, // ne
  2, // eq
  1, // ~
};
const int op_type[] = {
  0,
  0,
  0,
  0,
  0,
  0,

  NUMCOMP_GT,
  NUMCOMP_LT,
  NUMCOMP_EQ,
  NUMCOMP_EQ,
  NUMCOMP_NE,
  NUMCOMP_NE,
  NUMCOMP_LTEQ,
  NUMCOMP_GTEQ,

  STRCOMP_LIKE,
  STRCOMP_NE,
  STRCOMP_EQ,
  STRCOMP_SOMETHING,
};
const size_t id_count = sizeof(ids)/sizeof(ids[0]);

void skip_ws(struct state_t* state)
{
  while( state->curstart != state->queryend && isspace(*(state->curstart)) ) ++(state->curstart);
  state->curend = state->curstart;
}

void search_end(struct state_t* state)
{
  while( state->curend != state->queryend && !isspace(*(state->curend)) ) ++(state->curend);
}

void next_sym(struct state_t* state)
{
  skip_ws(state);
  search_end(state);
}

BOOL compare_id(const id_type id, struct state_t* state)
{
  next_sym(state);
  return ( state->curend - state->curstart == (int)id_lengths[id] ) && ( strncasecmp(state->curstart, ids[id], state->curend - state->curstart) == 0 );
}

BOOL accept_id(const id_type id, struct state_t* state)
{
  const BOOL retval = compare_id(id, state);
  state->curstart = state->curend;
  return retval;
}

id_type cursym_id(struct state_t* state)
{
  size_t i;
  for(i = 0; i < id_count; ++i) {
    if ( state->curend - state->curstart == (int)id_lengths[i]
      && strncasecmp(state->curstart, ids[i], state->curend - state->curstart) == 0 ) {
      return (id_type)i;
    }
  }
  return ID_NONE;
}

BOOL is_delimiter(const char c)
{
  static const char delims[] =
  {
    '(', ')', '>', '<', '=', '!', '~'
  };
  static const size_t delims_length = sizeof(delims)/sizeof(delims[0]);
  if ( isspace(c) ) return TRUE;
  {
    size_t i;
    for(i = 0; i < delims_length; ++i) {
      if ( c == delims[i] ) return TRUE;
    }
  }
  return FALSE;
}

void unquote(char* str)
{
  BOOL quoted = FALSE;
  BOOL backslash = FALSE;
  char* out = str;
  if ( !str ) return;
  while(*str) {
    if ( backslash ) {
      backslash = FALSE;
      *out = *str;
      ++out;
    }
    else if ( *str == '\\' ) {
      backslash = TRUE;
    }
    else if ( *str == '\'' ) {
      quoted = !quoted;
    }
    else {
      *out = *str;
      ++out;
    }
    ++str;
  }
  *out = '\0';
}

char* parse_textliteral(struct state_t* state)
{
  BOOL quoted = FALSE;
  BOOL backslash = FALSE;
  skip_ws(state);
  while( state->curend != state->queryend && (quoted || backslash || !is_delimiter(*(state->curend)) ) ) {
    if ( backslash ) {
      backslash = FALSE;
    }
    else if ( *(state->curend) == '\\' ) {
      backslash = TRUE;
    }
    else if ( *(state->curend) == '\'' ) {
      quoted = !quoted;
    }
    ++(state->curend);
  }
  {
    char* retval = duplicateStr(state->curstart, state->curend);
    state->curstart = state->curend;
    unquote(retval);
    return retval;
  }
}

char* parse_uri(struct state_t* state)
{
  skip_ws(state);
  {
    if ( state->curstart == state->queryend || *(state->curstart) != '<' ) return 0;
    ++(state->curstart);
    do {
      ++(state->curend);
    } while( state->curend != state->queryend && *(state->curend) != '>' );
    if ( state->curend == state->queryend ) return 0;
  }
  {
    char* retval = duplicateStr(state->curstart, state->curend);
    ++(state->curend); // this is needed to skip the last '>'
    state->curstart = state->curend;
    return retval;
  }
}

BOOL parse_numliteral(int* val, struct state_t* state)
{
  if ( val == 0 ) return FALSE;
  *val = 0;
  skip_ws(state);
  while( state->curend != state->queryend && !is_delimiter(*(state->curend)) ) {
    if ( *(state->curend) < '0' || *(state->curend) > '9' ) return FALSE;
    *val *= 10;
    *val += *(state->curend) - '0';
    ++(state->curend);
  }
  state->curstart = state->curend;
  return TRUE;
}

var_t parse_var(struct state_t* state)
{
  skip_ws(state);
  if ( state->curstart == state->queryend || *(state->curstart) != '?' ) return 0;
  ++(state->curstart);
  do {
    ++(state->curend);
  } while( state->curend != state->queryend && !is_delimiter(*(state->curend)) );
  {
    char* retval = duplicateStr(state->curstart, state->curend);
    state->curstart = state->curend;
    return retval;
  }
}

struct var_list_t* parse_var_list(struct state_t* state)
{
  struct var_list_t* first = 0;
  struct var_list_t* current = first;
  struct var_list_t* next = 0;

  while(TRUE) {
    var_t var = parse_var(state);
    if ( var == 0 ) break;
    next = (struct var_list_t*)getMem(sizeof(struct var_list_t));
    if ( next == 0 ) goto clean;
    next->var = var;
    if ( current ) {
      current->next = next;
      current = next;
    }
    else {
      first = current = next;
    }
  }

  return first;

clean:
  destroy_var_list(first);
  return 0;
}

BOOL parse_var_or_literal(struct var_or_literal_t* vol, struct state_t* state)
{
  if ( vol == 0 ) return FALSE;

  skip_ws(state);
  if ( state->curstart == state->queryend ) return FALSE;
  switch(*(state->curstart)) {
  case '?':
    vol->type = VAR_T;
    vol->text = parse_var(state);
    if ( vol->text == 0 ) return FALSE;
    break;
  case '<':
    vol->type = URI_T;
    vol->text = parse_uri(state);
    if ( vol->text == 0 ) return FALSE;
    break;
  default:
    if ( *(state->curstart) >= '0' && *(state->curstart) <= '9' ) {
      vol->type = NUMLITERAL_T;
      if ( !parse_numliteral(&(vol->num), state) ) return FALSE;
    }
    else {
      vol->type = TEXTLITERAL_T;
      vol->text = parse_textliteral(state);
      if ( vol->text == 0 ) return FALSE;
    }
  }

  return TRUE;
}

BOOL parse_triple_pattern(struct triple_pattern_t* tp, struct state_t* state)
{
  if ( tp == 0 ) return FALSE;

  skip_ws(state);
  if ( state->curstart == state->queryend ) return FALSE;
  if ( *(state->curstart) != '(' ) return FALSE;
  state->curstart++;
  state->curend = state->curstart;

  if ( !parse_var_or_literal(&(tp->subject), state) ) return FALSE;

  if ( !parse_var_or_literal(&(tp->predicate), state) ) return FALSE;

  if ( !parse_var_or_literal(&(tp->object), state) ) return FALSE;

  skip_ws(state);
  if ( state->curstart == state->queryend ) return FALSE;
  if ( *(state->curstart) != ')' ) return FALSE;
  state->curstart++;
  state->curend = state->curstart;

  return TRUE;
}

struct triple_pattern_list_t* parse_triple_pattern_list(struct state_t* state)
{
  struct triple_pattern_list_t* first = 0;
  struct triple_pattern_list_t* current = first;
  struct triple_pattern_list_t* next = 0;

  while(TRUE) {
    next = (struct triple_pattern_list_t*)getMem(sizeof(struct triple_pattern_list_t));
    if ( next == 0 ) goto clean;
    if ( !parse_triple_pattern(&(next->tp), state) ) {
      freeMem((char*)next);
      break;
    }
    if ( current ) {
      current->next = next;
      current = next;
    }
    else {
      first = current = next;
    }
  }

  return first;

clean:
  destroy_triple_pattern_list(first);
  return 0;
}

BOOL parse_op(int* op, struct state_t* state)
{
  int id;
  if ( op == 0 ) return FALSE;
  next_sym(state);
  id = cursym_id(state);
  if ( id == ID_NONE ) return FALSE;
  *op = op_type[id];
  if ( *op == 0 ) return FALSE;
  state->curstart = state->curend;
  return TRUE;
}

BOOL parse_expression(struct expression_t* expr, struct state_t* state)
{
  if ( expr == 0 ) return FALSE;

  expr->var = parse_var(state);
  if ( expr->var == 0 ) return FALSE;

  if ( !parse_op(&(expr->operation), state) ) return FALSE;

  if ( !parse_var_or_literal(&(expr->var_or_literal), state) ) return FALSE;

  if ( expr->operation < NUMCOMP && expr->var_or_literal.type == NUMLITERAL_T ) return FALSE;

  if ( expr->operation == STRCOMP_LIKE ) {
    expr->rx = regcomp(expr->var_or_literal.text);
    if ( !expr->rx ) return FALSE;
  }

  return TRUE;
}

struct constraint_list_t* parse_constraint_list(struct state_t* state)
{
  struct constraint_list_t* first = 0;
  struct constraint_list_t* current = first;
  struct constraint_list_t* next = 0;

  while(TRUE) {
    next = (struct constraint_list_t*)getMem(sizeof(struct constraint_list_t));
    if ( next == 0 ) goto clean;
    if ( !parse_expression(&(next->expr), state) ) {
      freeMem((char*)next);
      break;
    }
    if ( current ) {
      current->next = next;
      current = next;
    }
    else {
      first = current = next;
    }
  }

  return first;

clean:
  destroy_constraint_list(first);
  return 0;
}

BOOL parse_for_expr(struct for_list_t* fl, struct state_t* state)
{
    next_sym(state);
    if ( state->curend <= state->curstart ) return FALSE;
    fl->prefix = duplicateStr(state->curstart, state->curend);
    if ( fl->prefix == 0 ) return FALSE;
    state->curstart = state->curend;
    
    if ( !accept_id(ID_FOR, state) ) return FALSE;

    fl->url = parse_uri(state);
    if ( fl->url == 0 ) return FALSE;

    return TRUE;
}

struct for_list_t* parse_for_list(struct state_t* state)
{
  struct for_list_t* first = 0;
  struct for_list_t* current = first;
  struct for_list_t* next = 0;

  while(TRUE) {
    next = (struct for_list_t*)getMem(sizeof(struct for_list_t));
    if ( next == 0 ) goto clean;
    if ( !parse_for_expr(next, state) ) {
      freeMem((char*)next);
      break;
    }
    if ( current ) {
      current->next = next;
      current = next;
    }
    else {
      first = current = next;
    }
  }

  return first;

clean:
  destroy_for_list(first);
  return 0;
}

struct select_query_t* parse_select_query(struct state_t* state)
{
  struct select_query_t* select_query;

  if ( !accept_id(ID_SELECT, state) ) return 0;
  select_query = (struct select_query_t*)getMem(sizeof(struct select_query_t)); // getMem also zeroes the data
  if ( select_query == 0 ) return 0;
  select_query->vars = parse_var_list(state);
  if ( select_query->vars == 0 ) goto clean;

  if ( !accept_id(ID_FROM, state) ) goto clean;
  select_query->db = parse_textliteral(state);
  if ( select_query->db == 0 ) goto clean;

  if ( !accept_id(ID_WHERE, state) ) goto clean;
  select_query->triple_pattern_list = parse_triple_pattern_list(state);
  if ( select_query->triple_pattern_list == 0 ) goto clean;

  if ( compare_id(ID_AND, state) ) {
    select_query->constraint_list = parse_constraint_list(state);
    if ( select_query->constraint_list == 0 ) goto clean;
  }
  if ( compare_id(ID_USING, state) ) {
    select_query->for_list = parse_for_list(state);
    if ( select_query->for_list == 0 ) goto clean;
  }

  skip_ws(state);
  if ( state->curstart != state->queryend ) {
    goto clean;
  }

  return select_query;

clean:
  destroy_select_query(select_query);
  return 0;
}

struct select_query_t* parseSelectQuery(const char* query)
{
  struct state_t state;
  state.querystart=query;
  state.queryend=query+strlen(query);
  state.curstart=state.querystart;
  state.curend=0;
  return parse_select_query(&state);
}

void destroy_select_query(struct select_query_t* sq)
{
  if ( sq ) {
    if ( sq->vars ) destroy_var_list(sq->vars);
    if ( sq->db ) freeMem(sq->db);
    if ( sq->triple_pattern_list ) destroy_triple_pattern_list(sq->triple_pattern_list);
    if ( sq->constraint_list ) destroy_constraint_list(sq->constraint_list);
    if ( sq->for_list ) destroy_for_list(sq->for_list);
    freeMem((char*)sq);
  }
}

void destroy_var_list(struct var_list_t* vl)
{
  struct var_list_t* vlnext;
  while(vl) {
    if ( vl->var ) freeMem(vl->var);
    vlnext = vl->next;
    freeMem((char*)vl);
    vl = vlnext;
  }
}

void destroy_triple_pattern_list(struct triple_pattern_list_t* tpl)
{
  struct triple_pattern_list_t* tplnext;
  while(tpl) {
    destroy_triple_pattern(&(tpl->tp));
    tpl = tpl->next;
    tplnext = tpl->next;
    freeMem((char*)tpl);
    tpl = tplnext;
  }
}

void destroy_triple_pattern(struct triple_pattern_t* tp)
{
  if ( tp ) {
    destroy_var_or_literal(&(tp->object));
    destroy_var_or_literal(&(tp->predicate));
    destroy_var_or_literal(&(tp->subject));
  }
}

void destroy_var_or_literal(struct var_or_literal_t* vol)
{
  if ( vol && vol->text ) freeMem(vol->text);
}

void destroy_constraint_list(struct constraint_list_t* cl)
{
  struct constraint_list_t* clnext;
  while(cl) {
    destroy_expression(&(cl->expr));
    clnext = cl->next;
    freeMem((char*)cl);
    cl = clnext;
  }
}

void destroy_expression(struct expression_t* expr)
{
  if ( expr ) {
    if ( expr->var ) freeMem(expr->var);
    if ( expr->rx ) free(expr->rx); // I am using free(...) here because expr->rx was allocated by the regexp library.
    destroy_var_or_literal(&(expr->var_or_literal));
  }
}

void destroy_for_list(struct for_list_t* fl)
{
  struct for_list_t* flnext;
  while(fl) {
    if ( fl->prefix ) freeMem(fl->prefix);
    if ( fl->url ) freeMem(fl->url);
    flnext = fl->next;
    freeMem((char*)fl);
    fl = flnext;
  }
}

char* duplicateStr(const char* start, const char* end)
{
  char* retval;
  if ( end < start ) return 0;
  retval = malloc(end-start+1);
  if ( retval == 0 ) return 0;
  memcpy(retval, start, end-start);
  retval[end-start] = 0;
  return retval;
}
